@@ -68,15 +68,41 @@ module Darcs.Util.Index( readIndex, updateIndexFrom, indexFormatValid
     where
 
 import Prelude hiding ( lookup, readFile, writeFile, filter, (<$>) )
-import Darcs.Util.ByteString ( readSegment )
+import Darcs.Util.ByteString ( readSegment, decodeLocale )
+import Darcs.Util.File ( getFileStatus )
+import Darcs.Util.Hash( sha256, rawHash )
 import Darcs.Util.Tree
-import Darcs.Util.Path hiding ( getCurrentDirectory )
+import Darcs.Util.Path
+    ( AnchoredPath(..)
+    , anchorPath
+    , anchoredRoot
+    , unsafeMakeName
+    , appendPath
+    , flatten
+    )
+import Control.Monad( when )
+import Control.Exception( catch, SomeException )
+import Control.Applicative( (<$>) )
+
+import qualified Data.ByteString as B
+import qualified Data.ByteString.Char8 as BC
+import Data.ByteString.Unsafe( unsafeHead, unsafeDrop )
+import Data.ByteString.Internal( toForeignPtr, fromForeignPtr, memcpy
+                               , nullForeignPtr, c2w )
+
+import Data.Bits( Bits )
+#ifdef BIGENDIAN
+import Data.Bits( (.&.), (.|.), shift, shiftL, rotateR )
+#endif
 import Data.Int( Int64, Int32 )
+import Data.IORef( )
+import Data.Maybe( fromJust, isJust, fromMaybe )
+
+import Foreign.Storable
+import Foreign.ForeignPtr( ForeignPtr, withForeignPtr, castForeignPtr )
+import Foreign.Ptr( Ptr, plusPtr )
 
-import Bundled.Posix( getFileStatusBS, modificationTime,
-                      getFileStatus, fileSize, fileExists, isDirectory )
 import System.IO.MMap( mmapFileForeignPtr, mmapFileByteString, Mode(..) )
-import System.IO( )
 import System.Directory( doesFileExist, getCurrentDirectory, doesDirectoryExist )
 #if mingw32_HOST_OS
 import System.Directory( renameFile )
@@ -84,39 +110,22 @@ import System.FilePath( (<.>) )
 #else
 import System.Directory( removeFile )
 #endif
-import System.FilePath( (</>) )
-import System.Posix.Types ( FileID )
 
-import Control.Monad( when )
-import Control.Exception( catch, SomeException )
-import Control.Applicative( (<$>) )
-
-import qualified Data.ByteString as BS
-import qualified Data.ByteString.Char8 as BSC
-import Data.ByteString.Unsafe( unsafeHead, unsafeDrop )
-import Data.ByteString.Internal( toForeignPtr, fromForeignPtr, memcpy
-                               , nullForeignPtr, c2w )
-
-import Data.IORef( )
-import Data.Maybe( fromJust, isJust, fromMaybe )
-import Data.Bits( Bits )
-#ifdef BIGENDIAN
-import Data.Bits( (.&.), (.|.), shift, shiftL, rotateR )
-#endif
-
-import Foreign.Storable
-import Foreign.ForeignPtr( ForeignPtr, withForeignPtr, castForeignPtr )
-import Foreign.Ptr( Ptr, plusPtr )
-
-import Darcs.Util.Hash( sha256, rawHash )
 #ifdef WIN32
 import System.Win32.File ( createFile, getFileInformationByHandle, BY_HANDLE_FILE_INFORMATION(..),
                            fILE_SHARE_NONE, fILE_FLAG_BACKUP_SEMANTICS,
                            gENERIC_NONE, oPEN_EXISTING, closeHandle )
 #else
-import System.PosixCompat ( fileID, getSymbolicLinkStatus )
+import qualified System.Posix.Files as F ( getSymbolicLinkStatus, fileID )
 #endif
 
+import System.FilePath ( (</>) )
+import qualified System.Posix.Files as F
+    ( modificationTime, fileSize, isDirectory
+    , FileStatus
+    )
+import System.Posix.Types ( FileID, EpochTime, FileOffset )
+
 --------------------------
 -- Indexed trees
 --
@@ -131,7 +140,7 @@ import System.PosixCompat ( fileID, getSymbolicLinkStatus )
 -- recursive Tree object, which is rather expensive... As a bonus, we can also
 -- efficiently implement subtree queries this way (cf. 'readIndex').
 data Item = Item { iBase :: !(Ptr ())
-                 , iHashAndDescriptor :: !BS.ByteString
+                 , iHashAndDescriptor :: !B.ByteString
                  } deriving Show
 
 size_magic :: Int
@@ -154,16 +163,18 @@ off_dsc = off_hash + size_hash
 
 itemAllocSize :: AnchoredPath -> Int
 itemAllocSize apath =
-    align 4 $ size_hash + size_size + size_aux + size_fileid + size_dsclen + 2 + BS.length (flatten apath)
+    align 4 $ size_hash + size_size + size_aux + size_fileid + size_dsclen + 2 + B.length (flatten apath)
 
 itemSize, itemNext :: Item -> Int
-itemSize i = size_size + size_aux + size_fileid + size_dsclen + (BS.length $ iHashAndDescriptor i)
+itemSize i = size_size + size_aux + size_fileid + size_dsclen + (B.length $ iHashAndDescriptor i)
 itemNext i = align 4 (itemSize i + 1)
 
-iPath, iHash, iDescriptor :: Item -> BS.ByteString
+iHash, iDescriptor :: Item -> B.ByteString
 iDescriptor = unsafeDrop size_hash . iHashAndDescriptor
-iPath = unsafeDrop 1 . iDescriptor
-iHash = BS.take size_hash . iHashAndDescriptor
+iHash = B.take size_hash . iHashAndDescriptor
+
+iPath :: Item -> FilePath
+iPath = decodeLocale . unsafeDrop 1 . iDescriptor
 
 iSize, iAux :: Item -> Ptr Int64
 iSize i = plusPtr (iBase i) off_size
@@ -183,15 +194,29 @@ xlatePeek64 = fmap xlate64 . peek
 xlatePoke64 :: (Storable a, Num a, Bits a) => Ptr a -> a -> IO ()
 xlatePoke64 ptr v = poke ptr (xlate64 v)
 
+type FileStatus = Maybe F.FileStatus
+
+modificationTime :: FileStatus -> EpochTime
+modificationTime = maybe 0 F.modificationTime
+
+fileSize :: FileStatus -> FileOffset
+fileSize = maybe 0 F.fileSize
+
+fileExists :: FileStatus -> Bool
+fileExists = maybe False (const True)
+
+isDirectory :: FileStatus -> Bool
+isDirectory = maybe False F.isDirectory
+
 -- | Lay out the basic index item structure in memory. The memory location is
 -- given by a ForeignPointer () and an offset. The path and type given are
 -- written out, and a corresponding Item is given back. The remaining bits of
 -- the item can be filled out using 'update'.
 createItem :: ItemType -> AnchoredPath -> ForeignPtr () -> Int -> IO Item
 createItem typ apath fp off =
- do let dsc = BS.concat [ BSC.singleton $ if typ == TreeType then 'D' else 'F'
+ do let dsc = B.concat [ BC.singleton $ if typ == TreeType then 'D' else 'F'
                         , flatten apath
-                        , BS.singleton 0 ]
+                        , B.singleton 0 ]
         (dsc_fp, dsc_start, dsc_len) = toForeignPtr dsc
     withForeignPtr fp $ \p ->
         withForeignPtr dsc_fp $ \dsc_p ->
@@ -221,7 +246,7 @@ peekItem fp off =
 -- when updating directory entries).
 updateItem :: Item -> Int64 -> Hash -> IO ()
 updateItem item _ NoHash =
-    fail $ "Index.update NoHash: " ++ BSC.unpack (iPath item)
+    fail $ "Index.update NoHash: " ++ iPath item
 updateItem item size hash =
     do xlatePoke64 (iSize item) size
        unsafePokeBS (iHash item) (rawHash hash)
@@ -241,9 +266,7 @@ iHash' i = SHA256 (iHash i)
 -- the index file. mmapIndex will grow the index if it is smaller than this.
 mmapIndex :: forall a. FilePath -> Int -> IO (ForeignPtr a, Int)
 mmapIndex indexpath req_size = do
-  exist <- doesFileExist indexpath
-  act_size <- fromIntegral `fmap` if exist then fileSize `fmap` getFileStatus indexpath
-                                           else return 0
+  act_size <- fromIntegral . fileSize <$> getFileStatus indexpath
   let size = case req_size > 0 of
         True -> req_size
         False | act_size >= size_magic -> act_size - size_magic
@@ -293,19 +316,19 @@ readItem index state = do
   return res'
 
 readDir :: Index -> State -> Item -> IO Result
-readDir index state item =
-    do following <- fromIntegral <$> xlatePeek64 (iAux item)
-       st <- getFileStatusBS (iPath item)
+readDir index state item = do
+       following <- fromIntegral <$> xlatePeek64 (iAux item)
+       st <- getFileStatus (iPath item)
        let exists = fileExists st && isDirectory st
        fileid <- fromIntegral <$> (xlatePeek64 $ iFileID item)
-       fileid' <- fromMaybe fileid <$> (getFileID' $ BSC.unpack $ iPath item)
+       fileid' <- fromMaybe fileid <$> (getFileID' $ iPath item)
        when (fileid == 0) $ updateFileID item fileid'
-       let name it dirlen = Name $ (BS.drop (dirlen + 1) $ iDescriptor it) -- FIXME MAGIC
-           namelength = (BS.length $ iDescriptor item) - (dirlength state)
+       let name it dirlen = unsafeMakeName $ (B.drop (dirlen + 1) $ iDescriptor it) -- FIXME MAGIC
+           namelength = (B.length $ iDescriptor item) - (dirlength state)
            myname = name item (dirlength state)
            substate = state { start = start state + itemNext item
                             , path = path state `appendPath` myname
-                            , dirlength = if myname == Name (BSC.singleton '.')
+                            , dirlength = if myname == unsafeMakeName (BC.singleton '.')
                                              then dirlength state
                                              else dirlength state + namelength }
 
@@ -325,7 +348,7 @@ readDir index state item =
 
        let we_changed = or [ changed x | (_, x) <- inferiors ] || nullleaf
            nullleaf = null inferiors && oldhash == nullsha
-           nullsha = SHA256 (BS.replicate 32 0)
+           nullsha = SHA256 (B.replicate 32 0)
            tree' = makeTree [ (n, fromJust $ treeitem s) | (n, s) <- inferiors, isJust $ treeitem s ]
            treehash = if we_changed then hashtree index tree' else oldhash
            tree = tree' { treeHash = treehash }
@@ -338,15 +361,15 @@ readDir index state item =
                        , resitem = item }
 
 readFile :: Index -> State -> Item -> IO Result
-readFile index state item =
-    do st <- getFileStatusBS (iPath item)
+readFile index state item = do
+       st <- getFileStatus (iPath item)
        mtime <- fromIntegral <$> (xlatePeek64 $ iAux item)
        size <- xlatePeek64 $ iSize item
        fileid <- fromIntegral <$> (xlatePeek64 $ iFileID item)
-       fileid' <- fromMaybe fileid <$> (getFileID' $ BSC.unpack $ iPath item)
+       fileid' <- fromMaybe fileid <$> (getFileID' $ iPath item)
        let mtime' = modificationTime st
            size' = fromIntegral $ fileSize st
-           readblob = readSegment (basedir index </> BSC.unpack (iPath item), Nothing)
+           readblob = readSegment (basedir index </> (iPath item), Nothing)
            exists = fileExists st && not (isDirectory st)
            we_changed = mtime /= mtime' || size /= size'
            hash = iHash' item
@@ -394,12 +417,12 @@ readDirFileIDs :: Index -> State -> Item -> IO ResultF
 readDirFileIDs index state item =
     do fileid <- fromIntegral <$> (xlatePeek64 $ iFileID item)
        following <- fromIntegral <$> xlatePeek64 (iAux item)
-       let name it dirlen = Name $ (BS.drop (dirlen + 1) $ iDescriptor it) -- FIXME MAGIC
-           namelength = (BS.length $ iDescriptor item) - (dirlength state)
+       let name it dirlen = unsafeMakeName $ (B.drop (dirlen + 1) $ iDescriptor it) -- FIXME MAGIC
+           namelength = (B.length $ iDescriptor item) - (dirlength state)
            myname = name item (dirlength state)
            substate = state { start = start state + itemNext item
                             , path = path state `appendPath` myname
-                            , dirlength = if myname == Name (BSC.singleton '.')
+                            , dirlength = if myname == unsafeMakeName (BC.singleton '.')
                                              then dirlength state
                                              else dirlength state + namelength }
            subs off | off < following = do
@@ -417,7 +440,7 @@ readDirFileIDs index state item =
 readFileFileID :: Index -> State -> Item -> IO ResultF
 readFileFileID _ state item =
     do fileid' <- fromIntegral <$> (xlatePeek64 $ iFileID item)
-       let name it dirlen = Name $ (BS.drop (dirlen + 1) $ iDescriptor it)
+       let name it dirlen = unsafeMakeName $ (B.drop (dirlen + 1) $ iDescriptor it)
            myname = name item (dirlength state)
        return $ ResultF { nextF = start state + itemNext item
                         , resitemF = item
@@ -448,11 +471,11 @@ readIndex indexpath ht = do
 formatIndex :: ForeignPtr () -> Tree IO -> Tree IO -> IO ()
 formatIndex mmap_ptr old reference =
     do _ <- create (SubTree reference) (AnchoredPath []) size_magic
-       unsafePokeBS magic (BSC.pack "HSI5")
+       unsafePokeBS magic (BC.pack "HSI5")
     where magic = fromForeignPtr (castForeignPtr mmap_ptr) 0 4
           create (File _) path' off =
                do i <- createItem BlobType path' mmap_ptr off
-                  let flatpath = BSC.unpack $ flatten path'
+                  let flatpath = anchorPath "" path'
                   case find old path' of
                     Nothing -> return ()
                     -- TODO calling getFileStatus here is both slightly
@@ -503,12 +526,13 @@ updateIndexFrom indexpath hashtree' ref =
 -- | Check that a given file is an index file with a format we can handle. You
 -- should remove and re-create the index whenever this is not true.
 indexFormatValid :: FilePath -> IO Bool
-indexFormatValid path' =
-    do magic <- mmapFileByteString path' (Just (0, size_magic))
-       return $ case BSC.unpack magic of
-                  "HSI5" -> True
-                  _ -> False
-    `catch` \(_::SomeException) -> return False
+indexFormatValid path' = do
+    v <- do magic <- mmapFileByteString path' (Just (0, size_magic))
+            return $ case BC.unpack magic of
+                       "HSI5" -> True
+                       _ -> False
+         `catch` \(_::SomeException) -> return False
+    return v
 
 instance FilterTree IndexM IO where
     filter _ EmptyIndex = EmptyIndex
@@ -530,13 +554,13 @@ getFileID' fp = do file_exists <- doesFileExist fp
                               closeHandle h
                               return fhnumber
 #else
-                      then (Just . fileID) <$> getSymbolicLinkStatus fp
+                      then (Just . F.fileID) <$> F.getSymbolicLinkStatus fp
 #endif
                       else return Nothing
 
 
 -- Wow, unsafe.
-unsafePokeBS :: BSC.ByteString -> BSC.ByteString -> IO ()
+unsafePokeBS :: BC.ByteString -> BC.ByteString -> IO ()
 unsafePokeBS to from =
     do let (fp_to, off_to, len_to) = toForeignPtr to
            (fp_from, off_from, len_from) = toForeignPtr from
