@@ -28,8 +28,6 @@ module Darcs.Util.Path
     , fn2fp
     , fn2ps
     , ps2fn
-    , niceps2fn
-    , fn2niceps
     , breakOnDir
     , normPath
     , ownName
@@ -37,6 +35,8 @@ module Darcs.Util.Path
     , movedirfilename
     , encodeWhite
     , decodeWhite
+    , encodeWhiteName
+    , decodeWhiteName
     , isParentOrEqOf
     -- * AbsolutePath
     , AbsolutePath
@@ -76,7 +76,9 @@ module Darcs.Util.Path
     -- * AnchoredPaths: relative paths within a Tree. All paths are
     -- anchored at a certain root (this is usually the Tree root). They are
     -- represented by a list of Names (these are just strict bytestrings).
-    , Name(..)
+    , Name
+    , unsafeMakeName
+    , eqAnycase
     , AnchoredPath(..)
     , anchoredRoot
     , appendPath
@@ -84,7 +86,7 @@ module Darcs.Util.Path
     , isPrefix
     , parent, parents, catPaths, flatten, makeName, appendToName
     -- * Unsafe AnchoredPath functions.
-    , floatBS, floatPath, replacePrefixPath ) where
+    , floatPath, replacePrefixPath ) where
 
 import Prelude ()
 import Darcs.Prelude
@@ -96,7 +98,7 @@ import Data.List
     , intersect
     , inits
     )
-import Data.Char ( isSpace, chr, ord )
+import Data.Char ( isSpace, chr, ord, toLower )
 import Control.Exception ( tryJust, bracket_ )
 import Control.Monad ( when )
 import System.IO.Error ( isDoesNotExistError )
@@ -109,17 +111,14 @@ import qualified System.FilePath as NativeFilePath ( takeFileName, takeDirectory
 import System.FilePath( (</>), splitDirectories, normalise, dropTrailingPathSeparator )
 import System.Posix.Files ( isDirectory, getSymbolicLinkStatus )
 
-import Darcs.Util.ByteString ( packStringToUTF8, unpackPSFromUTF8 )
+import Darcs.Util.ByteString ( encodeLocale, decodeLocale )
 import qualified Data.ByteString.Char8 as BC
-import qualified Data.ByteString       as B  (ByteString)
+import qualified Data.ByteString       as B
 
 import Data.Binary
 import Darcs.Util.Global ( darcsdir )
 import Darcs.Util.URL ( isAbsolute, isRelative, isSshNopath )
 
-#include "impossible.h"
-
-
 -- | FileName is an abstract type intended to facilitate the input and output of
 -- unicode filenames.
 newtype FileName = FN FilePath deriving ( Eq, Ord )
@@ -140,21 +139,13 @@ fp2fn = FN
 fn2fp :: FileName -> FilePath
 fn2fp (FN fp) = fp
 
-{-# INLINE niceps2fn #-}
-niceps2fn :: B.ByteString -> FileName
-niceps2fn = FN . decodeWhite . BC.unpack
-
-{-# INLINE fn2niceps #-}
-fn2niceps :: FileName -> B.ByteString
-fn2niceps (FN fp) = BC.pack $ encodeWhite fp
-
 {-# INLINE fn2ps #-}
 fn2ps :: FileName -> B.ByteString
-fn2ps (FN fp) = packStringToUTF8 $ encodeWhite fp
+fn2ps (FN fp) = encodeLocale $ encodeWhite fp
 
 {-# INLINE ps2fn #-}
 ps2fn :: B.ByteString -> FileName
-ps2fn ps = FN $ decodeWhite $ unpackPSFromUTF8 ps
+ps2fn ps = FN $ decodeWhite $ decodeLocale ps
 
 {-# INLINE sp2fn #-}
 sp2fn :: SubPath -> FileName
@@ -190,6 +181,12 @@ decodeWhite cs_ = go cs_ [] False
            _ -> error "malformed filename"
        go (c:cs) acc modified = go cs (c:acc) modified
 
+encodeWhiteName :: Name -> B.ByteString
+encodeWhiteName = encodeLocale . encodeWhite . decodeLocale . unName
+
+decodeWhiteName :: B.ByteString -> Name
+decodeWhiteName = Name . encodeLocale . decodeWhite . decodeLocale
+
 ownName :: FileName -> FileName
 ownName (FN f) =  case breakLast '/' f of Nothing -> FN f
                                           Just (_,f') -> FN f'
@@ -568,7 +565,7 @@ floatSubPath = floatPath . fn2fp . sp2fn
 -- AnchoredPath utilities
 --
 
-newtype Name = Name BC.ByteString  deriving (Eq, Show, Ord)
+newtype Name = Name { unName :: B.ByteString } deriving (Eq, Show, Ord)
 
 -- | This is a type of "sane" file paths. These are always canonic in the sense
 -- that there are no stray slashes, no ".." components and similar. They are
@@ -586,7 +583,7 @@ isPrefix :: AnchoredPath -> AnchoredPath -> Bool
 appendPath :: AnchoredPath -> Name -> AnchoredPath
 appendPath (AnchoredPath p) n =
     case n of
-      (Name s) | s == BC.empty -> AnchoredPath p
+      (Name s) | B.null s -> AnchoredPath p
                | s == BC.pack "." -> AnchoredPath p
                | otherwise -> AnchoredPath $ p ++ [n]
 
@@ -608,14 +605,9 @@ parents (AnchoredPath x) = map AnchoredPath . init . inits $ x
 -- 'FilePath'. Moreover, you can use @anchorPath \"\"@ to get a relative
 -- 'FilePath'.
 anchorPath :: FilePath -> AnchoredPath -> FilePath
-anchorPath dir p = dir </> BC.unpack (flatten p)
+anchorPath dir p = dir </> decodeLocale (flatten p)
 {-# INLINE anchorPath #-}
 
--- | Unsafe. Only ever use on bytestrings that came from flatten on a
--- pre-existing AnchoredPath.
-floatBS :: BC.ByteString -> AnchoredPath
-floatBS = AnchoredPath . map Name . takeWhile (not . BC.null) . BC.split '/'
-
 flatten :: AnchoredPath -> BC.ByteString
 flatten (AnchoredPath []) = BC.singleton '.'
 flatten (AnchoredPath p) = BC.intercalate (BC.singleton '/')
@@ -624,7 +616,7 @@ flatten (AnchoredPath p) = BC.intercalate (BC.singleton '/')
 makeName :: String -> Name
 makeName ".." = error ".. is not a valid AnchoredPath component name"
 makeName n | '/' `elem` n = error "/ may not occur in a valid AnchoredPath component name"
-           | otherwise = Name $ BC.pack n
+           | otherwise = Name $ encodeLocale n
 
 -- | Take a relative FilePath and turn it into an AnchoredPath. The operation
 -- is (relatively) unsafe. Basically, by using floatPath, you are testifying
@@ -637,7 +629,7 @@ makeName n | '/' `elem` n = error "/ may not occur in a valid AnchoredPath compo
 floatPath :: FilePath -> AnchoredPath
 floatPath = make . splitDirectories . normalise . dropTrailingPathSeparator
   where make ["."] = AnchoredPath []
-        make x = AnchoredPath $ map (Name . BC.pack) x
+        make x = AnchoredPath $ map (Name . encodeLocale) x
 
 
 anchoredRoot :: AnchoredPath
@@ -653,11 +645,17 @@ replacePrefixPath (AnchoredPath (r:p)) b (AnchoredPath (r':p'))
     | otherwise = AnchoredPath []
 replacePrefixPath _ _ _ = AnchoredPath []
 
--- | Append a ByteString to the last Name of an AnchoredPath.
+-- | Append a String to the last Name of an AnchoredPath.
 appendToName :: AnchoredPath -> String -> AnchoredPath
 appendToName (AnchoredPath p) s = AnchoredPath (init p++[Name finalname])
-    where suffix = BC.pack s
+    where suffix = encodeLocale s
           finalname | suffix `elem` (BC.tails lastname) = lastname
                     | otherwise = BC.append lastname suffix
           lastname = case last p of
                         Name name -> name
+
+unsafeMakeName :: B.ByteString -> Name
+unsafeMakeName = Name
+
+eqAnycase :: Name -> Name -> Bool
+eqAnycase (Name a) (Name b) = BC.map toLower a == BC.map toLower b
